import unittest
from unittest import skip
from maxar_canvas_grid.grid import Cell, covers, Grid
import shapely
from shapely.geometry import shape, LineString, Point, box, GeometryCollection
from shapely.affinity import translate
 
class test_init(unittest.TestCase):
 
    def test_init_props(self):
        tile = Cell(2010, 1196, 12, zone=31)
        self.assertEqual(31, tile.zone)
        self.assertEqual('EPSG:32631', tile.proj)
        self.assertEqual('N', tile.hemisphere)
        self.assertEqual('Z31-031131213210', tile.id)

    def test_comparisons(self):
        tile_1 = Cell(2010, 1196, 12, zone=31)
        tile_2 = Cell(2010, 1196, 12, zone=31)
        tile_3 = Cell(2010, 1196, 12, zone=32)
        tile_4 = Cell(1006, 1196, 12, zone=31)
        self.assertEqual(tile_1, tile_2)
        self.assertNotEqual(tile_1, tile_3)
        self.assertNotEqual(tile_1, tile_4)
 
    def test_geom(self):
        tile = Cell(2010, 1196, 12, zone=31)
        bounds = tile.bounds
        self.assertEqual(bounds, [310000.0, 4255000.0, 315000.0, 4260000.0])
        self.assertEqual(bounds[2]-bounds[0], 5_000)
        self.assertEqual(bounds[3]-bounds[1], 5_000)
        self.assertEqual(type(shape(tile)), shapely.geometry.polygon.Polygon)

    def test_roundtrip(self):
        tile = Cell('Z31-031131213210')
        #test WGS84 currently
        point = tile.geom_WGS84.centroid
        results = list(covers(point))
        self.assertEqual(results[0], tile)

    def test_roundtrip_proj(self):
        tile = Cell('Z31-031131213210')
        #test WGS84 currently
        point = box(*tile.bounds).centroid
        results = list(covers(point, src_proj=tile.proj))
        self.assertEqual(results[0], tile)

    def test_bad_inits(self):
        # TODO:out of range
        #   tile = Cell(10**10, 10**10)
        # no zone
        with self.assertRaises(Exception) as e:
            tile = Cell(15, 56)
        # bad zone
        with self.assertRaises(Exception) as e:
            tile = Cell(15, 56, 1234)
        # bad ID 
        with self.assertRaises(Exception) as e:
            tile = Cell('panda')

    def test_init_from_quadkey(self):
        tile = Cell('031131213210', zone=14)
        self.assertEqual(14, tile.zone)
        self.assertEqual('Z14-031131213210', tile.id)

    def test_init_from_id(self):
        tile = Cell('Z31-031131213210')
        self.assertEqual(31, tile.zone)
        self.assertEqual('031131213210', tile.quadkey)
        self.assertEqual(2010, tile.x)
        self.assertEqual(1196, tile.y)
        self.assertEqual(12, tile.z)

    def test_contains(self):
        cell = Cell('Z31-03113121321')
        child = cell.children[0]
        not_child = Cell('Z31-031131213220')
        self.assertTrue(child in cell)
        self.assertFalse(not_child in cell)

    def test_family(self):
        cell = Cell('Z31-03113121321')
        # parents
        parent = cell.parent
        self.assertEqual(parent.zoom, cell.zoom-1)
        parent = cell.get_parent_at_zoom(cell.zoom-1)
        self.assertEqual(parent.zoom, cell.zoom-1)
        ggparent = cell.get_parent_at_zoom(cell.zoom-3)
        self.assertEqual(ggparent.zoom, cell.zoom-3)

        # parent/child relationships, roundtripping
        me = cell.get_parent_at_zoom(cell.zoom)
        self.assertEqual(me, cell)
        me = cell.children[0].parent
        self.assertEqual(me, cell)

        # get parent that isn't should error
        with self.assertRaises(Exception) as e:
            cell.get_parent_at_zoom(cell.zoom+1)
        with self.assertRaises(Exception) as e:
            cell.get_parent_at_zoom(cell.zoom+3)



class test_covers_geoms(unittest.TestCase):

    def test_covers_point(self):
        tile = Cell(2010, 1196, 12, zone=31)
        point = shape(tile).centroid
        results = list(Grid(tile.zone).covers(point, 12))
        self.assertEqual(len(results), 1)
        result = results[0]
        self.assertEqual(tile.zone, result.zone)
        self.assertEqual(tile.x, result.x)
        self.assertEqual(tile.y, result.y)

    def test_covers_line(self):
        tile = Cell(2010, 1196, 12, zone=31)
        point = shape(tile).centroid
        point2 = translate(point, 0, 7000, 0)
        line = LineString([point, point2])
        grid = Grid(31)
        results = list(grid.covers(line, 12))
        self.assertEqual(len(results), 2)
        result = results[0]
        self.assertEqual(tile.zone, result.zone)
        self.assertEqual(tile.x, result.x)
        self.assertIn(result.y, (tile.y, tile.y-1))

    def test_covers_geometrycollection(self):
        tile = Cell(2010, 1196, 12, zone=31)
        point = shape(tile).centroid
        circle1 = translate(point, 0, 100, 0).buffer(20)
        circle2 = translate(point, 0, 7000, 0).buffer(30)
        circle3 = translate(point, 100, 0, 0).buffer(40)
        collection = GeometryCollection((circle1, circle2, circle3))
        grid = Grid(31)
        results = list(grid.covers(collection, 12))
        self.assertEqual(len(results), 2)
        result = results[0]
        self.assertEqual(tile.zone, result.zone)
        self.assertEqual(tile.x, result.x)
        self.assertIn(result.y, (tile.y, tile.y-1))


    def test_covers_box(self):
        from shapely.affinity import translate
        tile = Cell(2010, 1196, 12, zone=31)
        box = translate(shape(tile), 2500, 2500, 0)
        grid = Grid(31)
        results = list(grid.covers(box, zoom=12))
        self.assertEqual(len(results), 4)


    def test_covers_geometrycollection_adjacent(self):
        tile = Cell(2010, 1196, 12, zone=31)
        point = shape(tile).centroid
        # make 2 boxes that are touching
        box1 = box(*point.buffer(20).bounds)
        box2 = translate(box1, 40, 0, 0)
        assert box1.touches(box2)
        assert box1.intersection(box2).area == 0
        collection = GeometryCollection((box1, box2))
        grid = Grid(31)
        results = list(grid.covers(collection, 12))
        self.assertEqual(len(results), 1)
        result = results[0]
        self.assertEqual(tile.zone, result.zone)
        self.assertEqual(tile.x, result.x)
        self.assertEqual(tile.y, result.y)
    

    def test_covers_multipolygon_split_across_antimeridian(self):

        # show poly with x values > 180 work across the antimeridian
        expected_cells = 1104
        input_geom = "POLYGON ((179 0, 179 1, 181 1, 181 0, 179 0))"
        results = list(covers(input_geom, zoom=12))
        all_count = len(results)
        #assert len(results) == expected_cells

        # get count for one side of the antimeridian
        expected_cells = 552
        input_geom = "POLYGON ((-180 0, -180 1, -179 1, -179 0, -180 0))"
        results = list(covers(input_geom, zoom=12))
        west_count = len(results)
        #assert len(results) == expected_cells

        # get count for the other side of the antimeridian
        expected_cells = 552
        input_geom = "POLYGON ((179 0, 179 1, 180 1, 180 0, 179 0))"
        results = list(covers(input_geom, zoom=12))
        east_count = len(results)
        #assert len(results) == expected_cells

        # test the combined multipolygon split across the antimeridian
        expected_cells = 1104  # covers() previously returned 552
        input_geom = "MULTIPOLYGON (((-180 0, -180 1, -179 1, -179 0, -180 0)), ((179 0, 179 1, 180 1, 180 0, 179 0)))"
        results = list(covers(input_geom, zoom=12))
        multi_count = len(results)
        #assert len(results) == expected_cells
        
        assert all_count == east_count + west_count
        assert all_count == multi_count

        # check there are no duplicate cells
        assert len(set([cell.id for cell in results])) == all_count

class test_grid_file_layout(unittest.TestCase):
    # from the Spaceknow bug

    def test_in_grid(self):
        g = Grid(36)
        c = Cell('Z36-013133013330')
        self.assertTrue(c in g)

        c = Cell('Z36-013133013312')
        self.assertTrue(c in g)

        c = Cell('Z36-013133013310')
        self.assertFalse(c in g)

        g = Grid(35)
        c = Cell('Z35-102022102201')
        self.assertTrue(c in g)

        g = Grid(35)
        c = Cell('Z35-102022100021')
        self.assertTrue(c in g)
        

    def test_stagger(self):

        # stagger in rows from grid files
        # checked in QGIS

        # original problem area
        cells = list(covers('POINT(30 60.02)'))
        self.assertEqual(len(cells), 1)

        cells = list(covers('POINT(30 59.97)'))
        self.assertEqual(len(cells), 2)
        
        # check again further south
        cells = list(covers('POINT(30 11.36)'))
        self.assertEqual(len(cells), 1)

        cells = list(covers('POINT(30 11.32)'))
        self.assertEqual(len(cells), 2)
        # check close to the equator

        cells = list(covers('POINT(30 0.1)'))
        self.assertEqual(len(cells), 2)

        #check on the equator (seam N/S)
        cells = list(covers('POINT(30 0)'))
        self.assertEqual(len(cells), 4)

        # check south of the equator
        cells = list(covers('POINT(30 -60.02)'))
        self.assertEqual(len(cells), 1)

        cells = list(covers('POINT(30 -59.97)'))
        self.assertEqual(len(cells), 2)
            
        # last stagger before falling off the top of the zone
        cells = list(covers('POINT(30 83.63)'))
        self.assertEqual(len(cells), 1)

        cells = list(covers('POINT(30 83.58)'))
        self.assertEqual(len(cells), 2)

class test_intersects_serialized(unittest.TestCase):
 
    def test_covers_point(self):
        point = 'POINT(58.0 15.95)'
        results = list(covers(point))
        self.assertEqual(len(results), 1)
        tile = results[0]
        self.assertEqual(tile.zone, 40)
        self.assertEqual(tile.hemisphere, 'N')
        self.assertEqual(tile.proj, 'EPSG:32640')
        self.assertEqual(tile.id, 'Z40-122020032323')

    def test_covers_point_southern(self):
        point = 'POINT(0.54 -26.78)'
        results = list(covers(point))
        self.assertEqual(len(results), 1)
        tile = results[0]
        self.assertEqual(tile.zone, 31)
        self.assertEqual(tile.hemisphere, 'S')
        self.assertEqual(tile.proj, 'EPSG:32731')
        self.assertEqual(tile.id, 'Z31-213113021111')

    def test_covers_point_with_proj(self):
        point = 'POINT(58.0 15.95)'
        results = list(covers(point, src_proj='EPSG:4326'))
        self.assertEqual(len(results), 1)
        tile = results[0]
        self.assertEqual(tile.id, 'Z40-122020032323')

    def test_covers_line(self):
        line = 'LINESTRING(58.0 15.95, 58.0 16.15)'
        results = list(covers(line))
        self.assertEqual(len(results), 6)

        actual_zones = set((40,))
        result_zones = set([tile.zone for tile in results])
        self.assertEqual(result_zones, actual_zones)

        result_keys = set([tile.quadkey for tile in results])
        actual_keys = set((
            '122020032321',
            '122020032301',
            '122020032303',
            '122020032123',
            '122020032323',
            '122020032121'
        ))    
        self.assertEqual(result_keys, actual_keys)
 
    def test_intersect_poly(self):
        poly = '{"type":"Polygon","coordinates":[[[58,15.95],[58,16.15],[58.08,16.15],[58.08,15.95],[58,15.95]]]}'
        results = list(covers(poly))
        self.assertEqual(len(results), 18)

        actual_zones = set((40,))
        result_zones = set([tile.zone for tile in results])
        self.assertEqual(result_zones, actual_zones)

        result_keys = set([tile.quadkey[:11] for tile in results])
        actual_keys = set(('12202003232', '12202003230', '12202003212',
                           '12202003233', '12202003231', '12202003213'))
        self.assertEqual(result_keys, actual_keys)

    def test_covers_python_geom(self):
        poly = {'type': 'Polygon', 'coordinates': [[[58, 15.95], [58, 16.15], [58.08, 16.15], [58.08, 15.95], [58, 15.95]]]}
        results = list(covers(poly))
        self.assertEqual(len(results), 18)

    def test_intersect_donut(self):
        center = Point(57, 22)
        outside = center.buffer(2)
        inside = center.buffer(1)
        donut = outside - inside 
        donut_results = list(covers(donut))
        outside_results = list(covers(outside))
        self.assertGreater(len(outside_results), len(donut_results))

    def test_crosses_zone(self):
        line = 'LINESTRING(59.95 22, 60.05 22)'
        results = list(covers(line))
        self.assertEqual(len(results), 4)
        actual_zones = set((40,41))
        result_zones = set([tile.zone for tile in results])
        self.assertEqual(actual_zones, result_zones)
    
    def test_near_zone(self):
        geocell = box(12, 47.5, 12.25, 47.75)
        results = list(covers(geocell))
        # value is from selecting in QGIS against grid files
        self.assertEqual(len(results), 35)
        actual_zones = set((32, 33))
        result_zones = set([tile.zone for tile in results])
        self.assertEqual(actual_zones, result_zones)

        overlapped_cell = 'z33-013333032233'
        results = list(covers(Cell(overlapped_cell).geom_WGS84.centroid))
        self.assertEqual(len(results), 2)

        solo_cell = 'z33-013333032322'
        results = list(covers(Cell(solo_cell).geom_WGS84.centroid))
        self.assertEqual(len(results), 1)

    def test_crosses_antimeridian(self):
        line = 'LINESTRING(179.95 40, 180.05 40)'
        results = list(covers(line))
        # note that though the test line is symmetrical around 180d
        # ARD grids are not symmetrical, so at this longitude the
        # grid coverage has two cells on the west side
        # and one cell on the east side
        self.assertEqual(len(results), 3)
        actual_zones = set((60,1))
        result_zones = set([tile.zone for tile in results])
        self.assertEqual(actual_zones, result_zones)

        

class test_geom(unittest.TestCase):
 
    def test_north(self):
        point = Point(-121, 40)
        tile = list(covers(point))[0]
        self.assertEqual(tile.hemisphere, 'N')
        self.assertEqual(tile.proj, 'EPSG:32610')
        self.assertTrue(shape(tile.geom_WGS84).contains(point))

    def test_south(self):
        point = Point(-121, -40)
        tile = list(covers(point))[0]
        self.assertEqual(tile.hemisphere, 'S')
        self.assertEqual(tile.proj, 'EPSG:32710')
        self.assertTrue(shape(tile.geom_WGS84).contains(point))


    
class test_neighbors(unittest.TestCase):
 
    def test_neighbors(self):
        '''
        cells go 

        0 1
        2 3

        so we can quickly check cells within a zone '''

        c = Cell('123123100', zone=11) 
        self.assertEqual(c.neighbor('E'), Cell('123123101', zone=11))
        self.assertEqual(c.neighbor('SE'), Cell('123123103', zone=11))

    def test_roundtrip(self):
        c = Cell('123123100', zone=11) 
        self.assertEqual(c, c.neighbor('N').neighbor('S'))
        self.assertEqual(c, c.neighbor('NE').neighbor('SW'))
        self.assertEqual(c, c.neighbor('NW').neighbor('SE'))
        self.assertEqual(c, c.neighbor('E').neighbor('W'))


if __name__ == '__main__':
    unittest.main()
